unit main;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, OpenGLContext, Forms, Controls, Graphics,
  Dialogs, ExtCtrls, StdCtrls, gl, glu, Types;

type
  TMatrix4f = array[0..3, 0..3] of GLfloat;

  { TForm1 }

  TForm1 = class(TForm)
    ApplicationProperties1: TApplicationProperties;
    Label1: TLabel;
    OpenGLControl1: TOpenGLControl;
    Panel1: TPanel;
    procedure ApplicationProperties1Idle(Sender: TObject; var Done: Boolean);
    procedure FormCreate(Sender: TObject);
    procedure OpenGLControl1MouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure OpenGLControl1MouseWheel(Sender: TObject; Shift: TShiftState;
      WheelDelta: Integer; MousePos: TPoint; var Handled: Boolean);
    procedure OpenGLControl1Paint(Sender: TObject);
  private
    { private declarations }
    MyMousePos: TPoint;
    MyMatrix: TMatrix4f;
  public
    { public declarations }
  end;

var
  Form1: TForm1;

implementation

{$R *.lfm}

uses
  Math;

procedure Identity(var m: TMatrix4f);
const
  mi: TMatrix4f = ((1, 0, 0, 0), (0, 1, 0, 0), (0, 0, 1, 0), (0, 0, 0, 1));
begin
  m := mi;
end;

procedure RotateA(var m: TMatrix4f; Winkel: GLfloat);
var
  i: integer;
  y, z, c, s: GLfloat;
begin
  c := cos(Winkel);
  s := sin(Winkel);
  for i := 0 to 2 do begin
    y := m[i, 1];
    z := m[i, 2];
    m[i, 1] := y * c - z * s;
    m[i, 2] := y * s + z * c;
  end;
end;

procedure RotateB(var m: TMatrix4f; Winkel: GLfloat);
var
  i: integer;
  x, z, c, s: GLfloat;
begin
  c := cos(Winkel);
  s := sin(Winkel);
  for i := 0 to 2 do begin
    x := m[i, 0];
    z := m[i, 2];
    m[i, 0] := x * c - z * s;
    m[i, 2] := x * s + z * c;
  end;
end;

procedure RotateC(var m: TMatrix4f; Winkel: GLfloat);
var
  i: integer;
  x, y, c, s: GLfloat;
begin
  c := cos(Winkel);
  s := sin(Winkel);
  for i := 0 to 2 do begin
    x := m[i, 0];
    y := m[i, 1];
    m[i, 0] := x * c - y * s;
    m[i, 1] := x * s + y * c;
  end;
end;

procedure Translate(var m: TMatrix4f; x, y, z: GLfloat);
begin
  m[3, 0] := m[3, 0] + x;
  m[3, 1] := m[3, 1] + y;
  m[3, 2] := m[3, 2] + z;
end;

procedure Scale(var m: TMatrix4f; Faktor: GLfloat);
var
  x, y: integer;
begin
  for x := 0 to 2 do begin
    for y := 0 to 2 do begin
      m[x, y] *= Faktor;
    end;
  end;
end;

{ TForm1 }

procedure TForm1.ApplicationProperties1Idle(Sender: TObject; var Done: Boolean);
begin
  Done := false;
  OpenGLControl1.Invalidate;
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
  Identity(MyMatrix);
end;

procedure TForm1.OpenGLControl1MouseMove(Sender: TObject; Shift: TShiftState;
  X, Y: integer);
const
  MoveStep = 1.0 / 100;
  RotStep = 100;
var
  phi1, phi2, dphi: Double;
  ctr: TPoint;
begin
  if Shift = [ssLeft] then begin
    Translate(MyMatrix, (X - MyMousePos.x) * MoveStep, -(Y - MyMousePos.y) * MoveStep, 0.0);
  end;
  if Shift = [ssRight] then begin
    RotateB(MyMatrix, -(X - MyMousePos.x) / RotStep);
    RotateA(MyMatrix, (Y - MyMousePos.y) / RotStep);
  end;
  if Shift = [ssMiddle] then begin
    ctr := Point(Width div 2, Height div 2);
    phi1 := arctan2(MyMousePos.Y - ctr.Y, MyMousePos.X - ctr.X);
    phi2 := arctan2(Y - ctr.Y, X - ctr.X);
    dphi := phi2 - phi1;
    if (phi2 > pi/2) and (phi1 < -pi/2) then
      dphi := 2*pi - dphi
    else
    if (phi1 > pi/2) and (phi2 < -pi/2) then
      dphi := 2*pi + dphi;
    RotateC(MyMatrix, -RadToDeg(dphi) / RotStep);
  end;

  MyMousePos.x := X;
  MyMousePos.y := Y;
end;

procedure TForm1.OpenGLControl1MouseWheel(Sender: TObject; Shift: TShiftState;
  WheelDelta: integer; MousePos: TPoint; var Handled: boolean);
const
  ZOOM_OUT = 1.05;
  ZOOM_IN = 1 / ZOOM_OUT;
begin
  glMatrixMode(GL_MODELVIEW);
  if WheelDelta > 0 then begin
    Scale(MyMatrix, ZOOM_IN);
  end else begin
    Scale(MyMatrix, ZOOM_OUT);
  end;
end;

var
  seed: Integer;

function randu: Double;
const
  imax = 2147483647;
begin
  seed := seed * 65539;
  if seed < 0 then seed := seed + imax + 1;
  result := seed / imax;
  seed := seed;
end;


procedure TForm1.OpenGLControl1Paint(Sender: TObject);
const
  N = 10000;
var
  i: Integer;
  x, y, z: Double;
begin
  glClearColor(0.0, 0.5, 1.0, 1.0);
  glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);
  glEnable(GL_DEPTH_TEST);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(10.0, double(OpenGLControl1.Width) / OpenGLControl1.Height, 0.1, 100.0);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity;
  glTranslatef(0, 0, -16);     // Anfangsabstand der Kamera
  glRotatef(7, 1,0,0);
  glRotatef(-12, 0,1,0);

  glMultMatrixf(@MyMatrix);

  glBegin(GL_LINES);
    glColor3f(1, 0, 0);
    glVertex3f(0, 0, 0);
    glVertex3f(10, 0, 0);
  glEnd;
  glBegin(GL_LINES);
    glColor3f(0, 1, 0);
    glVertex3f(0, 0, 0);
    glVertex3f(0, 10, 0);
  glEnd;
  glBegin(GL_LINES);
    glColor3f(0, 0, 1);
    glVertex3f(0, 0, 0);
    glVertex3f(0, 0, 10);
  glEnd;

  seed := 1;

  for i := 0 to N do begin
    x := randu;
    y := randu;
    z := randu;
    glBegin(GL_POINTS);
       glColor3f(1,1,1);
       glVertex3f(x, y, z);
    glEnd();
  end;

  OpenGLControl1.SwapBuffers;
end;

end.

